@@ -294,6 +294,8 @@ void dump_config(struct runctl *runp, struct query *querylist)
 		stringdump("auth", "kerberos_v5");
 	    else if (ctl->server.authenticate == A_SSH)
 		stringdump("auth", "ssh");
+	    else if (ctl->server.authenticate == A_OTP)
+		stringdump("auth", "otp");
 
 #if defined(HAVE_GETHOSTBYNAME) && defined(HAVE_RES_SEARCH)
 	    booldump("dns", ctl->server.dns);
@@ -311,6 +313,10 @@ void dump_config(struct runctl *runp, struct query *querylist)
 	    stringdump("plugin", ctl->server.plugin);
 	    stringdump("plugout", ctl->server.plugout);
 	    stringdump("principal", ctl->server.principal);
+	    if (ctl->server.esmtp_name)
+	        stringdump("esmtpname",ctl->server.esmtp_name);
+	    if (ctl->server.esmtp_password)
+	        stringdump("esmtppassword",ctl->server.esmtp_password);
 	    booldump("tracepolls", ctl->tracepolls);
 
 	    indent(0);
@@ -370,6 +376,8 @@ void dump_config(struct runctl *runp, struct query *querylist)
 	numdump("limit", ctl->limit);
 	numdump("warnings", ctl->warnings);
 	numdump("fetchlimit", ctl->fetchlimit);
+	numdump("fetchsizelimit", ctl->fetchsizelimit);
+	numdump("fastuidl", ctl->fastuidl);
 	numdump("batchlimit", ctl->batchlimit);
 #ifdef SSL_ENABLE
 	booldump("ssl", ctl->use_ssl);
@@ -389,18 +397,13 @@ void dump_config(struct runctl *runp, struct query *querylist)
 
 	indent('\0');
 	fprintf(stdout, "'antispam':'");
-	if (!ctl->antispam)
-	    fputs("'\n", stdout);
-	else
+	for (idp = ctl->antispam; idp; idp = idp->next)
 	{
-	    for (idp = ctl->antispam; idp; idp = idp->next)
-	    {
-		fprintf(stdout, "%d", idp->val.status.num);
-		if (idp->next)
-		    fputs(" ", stdout);
-	    }
-	    fputs("',\n", stdout);
+	    fprintf(stdout, "%d", idp->val.status.num);
+	    if (idp->next)
+		fputs(" ", stdout);
 	}
+	fputs("',\n", stdout);
 	listdump("mailboxes", ctl->mailboxes);
 
 	indent('}');
