@@ -66,6 +66,18 @@ void envquery(int argc, char **argv)
 	exit(PS_UNDEFINED);
     }
 
+    if (getenv("NULLMAILER_FLAGS") && strcmp(getenv("NULLMAILER_FLAGS"), ""))
+    {
+	fprintf(stderr,
+		GT_("%s: The NULLMAILER_FLAGS environment variable is set.\n"
+		    "This is dangerous as it can make nullmailer-inject or nullmailer's\n" 
+		    "sendmail wrapper tamper with your From:, Message-ID: or Return-Path: headers.\n"
+		    "Try \"env NULLMAILER_FLAGS= %s YOUR ARGUMENTS HERE\"\n"
+		    "%s: Abort.\n"), 
+		program_name, program_name, program_name);
+	exit(PS_UNDEFINED);
+    }
+
     if (!(pwp = getpwuid(getuid())))
     {
 	fprintf(stderr,
@@ -201,12 +213,12 @@ char *rfc822timestamp(void)
      * weird multibyte i18n characters (such as kanji) from showing up
      * in your Received headers.
      */
-#if defined(HAVE_SETLOCALE) && defined(ENABLE_NLS) && defined(HAVE_STRFTIME)
+#if defined(HAVE_SETLOCALE) && defined(ENABLE_NLS)
     setlocale (LC_TIME, "C");
 #endif
     strftime(buf, sizeof(buf)-1, 
 	     "%a, %d %b %Y %H:%M:%S XXXXX (%Z)", localtime(&now));
-#if defined(HAVE_SETLOCALE) && defined(ENABLE_NLS) && defined(HAVE_STRFTIME)
+#if defined(HAVE_SETLOCALE) && defined(ENABLE_NLS)
     setlocale (LC_TIME, "");
 #endif
     strncpy(strstr(buf, "XXXXX"), tzoffset(&now), 5);
