@@ -86,6 +86,8 @@
 #define LA_SSLFINGERPRINT	60
 #endif
 
+#define LA_FETCHSIZELIMIT	61
+#define LA_FASTUIDL	62
 
 /* options still left: CDgGhHjJoORwWxXYz */
 static const char *shortoptions = 
@@ -140,6 +142,8 @@ static const struct option longoptions[] = {
   
   {"batchlimit",required_argument, (int *) 0, LA_BATCHLIMIT  },
   {"fetchlimit",required_argument, (int *) 0, LA_FETCHLIMIT  },
+  {"fetchsizelimit",required_argument, (int *) 0, LA_FETCHSIZELIMIT  },
+  {"fastuidl",	required_argument, (int *) 0, LA_FASTUIDL    },
   {"expunge",	required_argument, (int *) 0, LA_EXPUNGE     },
   {"mda",	required_argument, (int *) 0, LA_MDA         },
   {"bsmtp",	required_argument, (int *) 0, LA_BSMTP       },
@@ -319,7 +323,7 @@ struct query *ctl;	/* option record to be initialized */
 	    rctl->invisible = TRUE;
 	    break;
 	case LA_SHOWDOTS:
-	    rctl->showdots = TRUE;
+	    rctl->showdots = FLAG_TRUE;
 	    break;
 	case 'f':
 	case LA_RCFILE:
@@ -409,6 +413,8 @@ struct query *ctl;	/* option record to be initialized */
 		ctl->server.authenticate = A_SSH;
 	    else if (strcmp(optarg, "otp") == 0)
 		ctl->server.authenticate = A_OTP;
+	    else if (strcmp(optarg, "opie") == 0)
+		ctl->server.authenticate = A_OTP;
 	    else if (strcmp(optarg, "ntlm") == 0)
 		ctl->server.authenticate = A_NTLM;
 	    else if (strcmp(optarg, "cram") == 0)
@@ -527,6 +533,14 @@ struct query *ctl;	/* option record to be initialized */
 	    c = xatoi(optarg, &errflag);
 	    ctl->fetchlimit = NUM_VALUE_IN(c);
 	    break;
+	case LA_FETCHSIZELIMIT:
+	    c = xatoi(optarg, &errflag);
+	    ctl->fetchsizelimit = NUM_VALUE_IN(c);
+	    break;
+	case LA_FASTUIDL:
+	    c = xatoi(optarg, &errflag);
+	    ctl->fastuidl = NUM_VALUE_IN(c);
+	    break;
 	case 'e':
 	case LA_EXPUNGE:
 	    c = xatoi(optarg, &errflag);
@@ -677,7 +691,7 @@ struct query *ctl;	/* option record to be initialized */
 	P(GT_("  -p, --protocol    specify retrieval protocol (see man page)\n"));
 	P(GT_("  -U, --uidl        force the use of UIDLs (pop3 only)\n"));
 	P(GT_("  -P, --port        TCP/IP service port to connect to\n"));
-	P(GT_("      --auth        authentication type (password/kerberos/ssh)\n"));
+	P(GT_("      --auth        authentication type (password/kerberos/ssh/otp)\n"));
 	P(GT_("  -t, --timeout     server nonresponse timeout\n"));
 	P(GT_("  -E, --envelope    envelope address header\n"));
 	P(GT_("  -Q, --qvirtual    prefix to remove from local user id\n"));
@@ -703,6 +717,8 @@ struct query *ctl;	/* option record to be initialized */
 	P(GT_("  -Z, --antispam,   set antispam response values\n"));
 	P(GT_("  -b, --batchlimit  set batch limit for SMTP connections\n"));
 	P(GT_("  -B, --fetchlimit  set fetch limit for server connections\n"));
+	P(GT_("      --fetchsizelimit set fetch message size limit\n"));
+	P(GT_("      --fastuidl    do a binary search for UIDLs\n"));
 	P(GT_("  -e, --expunge     set max deletions between expunges\n"));
         P(GT_("  -m, --mda         set MDA to use for forwarding\n"));
         P(GT_("      --bsmtp       set output BSMTP file\n"));
