@@ -73,6 +73,11 @@ int smtp_open(struct query *ctl)
     char *parsed_host = NULL;
 
     /* maybe it's time to close the socket in order to force delivery */
+    if (last_smtp_ok > 0 && time((time_t *)NULL) - last_smtp_ok > mytimeout)
+    {
+	smtp_close(ctl, 1);
+	last_smtp_ok = 0;
+    }
     if (NUM_NONZERO(ctl->batchlimit)) {
 	if (batchcount == ctl->batchlimit)
 	    smtp_close(ctl, 1);
@@ -206,7 +211,39 @@ int smtp_open(struct query *ctl)
 		ctl->destaddr = (ctl->smtphost && ctl->smtphost[0] != '/') ? ctl->smtphost : "localhost";
     } 
     else 
-	ctl->destaddr = ctl->smtpaddress ? ctl->smtpaddress : ( ctl->smtphost && ctl->smtphost[0] != '/' ? ctl->smtphost : "localhost");
+      {
+	/* 
+	 * Here we try to find a correct domain name part for the RCPT
+	 * TO address.  If smtpaddress is set, no need to guestimate
+	 * it.  Otherwise, using ctl->smtphost as a base is a good
+	 * base, although we may have to strip any port appended to
+	 * communicate with SMTP servers that do not listen on the
+	 * SMTP port.  (benj) */
+	if (ctl->smtpaddress)
+	  ctl->destaddr = ctl->smtpaddress;
+	else if (ctl->smtphost && ctl->smtphost[0] != '/')
+	  {
+	    char * cp;
+	    if (cp = strchr (ctl->smtphost, '/'))
+	    {
+	      /* As an alternate port for smtphost is specified, we
+		 need to strip it from domain name. */
+	      char *smtpname;
+	      xalloca(smtpname, char *, cp - ctl->smtphost + 1);
+	      strncpy(smtpname, ctl->smtphost, cp - ctl->smtphost +1);
+	      cp = strchr(smtpname, '/');
+	      *cp = 0;
+	      ctl->destaddr = smtpname;
+	    }
+	    else
+	      /* No need to strip port, domain name is smtphost. */
+	      ctl->destaddr = ctl->smtphost;
+	  }
+	/* No smtphost is specified or it is a UNIX socket, then use
+	   localhost as a domain part. */
+	else
+	  ctl->destaddr = "localhost";
+      }
 
     if (outlevel >= O_DEBUG && ctl->smtp_socket != -1)
 	report(stdout, GT_("forwarding to %s\n"), ctl->smtphost);
@@ -215,12 +252,11 @@ int smtp_open(struct query *ctl)
 }
 
 static void sanitize(char *s)
-/* replace unsafe shellchars by an _ */
+/* replace ' by _ */
 {
-    const static char *ok_chars = " 1234567890!@%-_=+:,./abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
     char *cp;
 
-    for (cp = s; *(cp += strspn(cp, ok_chars)); /* NO INCREMENT */)
+    for (cp = s; (cp = strchr (cp, '\'')); cp++)
     	*cp = '_';
 }
 
@@ -263,17 +299,24 @@ static int send_bouncemail(struct query *ctl, struct msgblk *msg,
     char daemon_name[18 + HOSTLEN] = "FETCHMAIL-DAEMON@";
     char boundary[BUFSIZ], *bounce_to;
     int sock;
+    static char *fqdn_of_host = NULL;
+    const char *md1 = "MAILER-DAEMON", *md2 = "MAILER-DAEMON@";
 
     /* don't bounce in reply to undeliverable bounces */
-    if (!msg->return_path[0] || strcmp(msg->return_path, "<>") == 0)
-	return(FALSE);
+    if (!msg->return_path[0] ||
+	strcmp(msg->return_path, "<>") == 0 ||
+	strcasecmp(msg->return_path, md1) == 0 ||
+	strncasecmp(msg->return_path, md2, strlen(md2)) == 0)
+	return(TRUE);
 
     bounce_to = (run.bouncemail ? msg->return_path : run.postmaster);
 
     SMTP_setmode(SMTP_MODE);
 
     /* can't just use fetchmailhost here, it might be localhost */
-    strcat(daemon_name, host_fqdn());
+    if (fqdn_of_host == NULL)
+	fqdn_of_host = host_fqdn();
+    strcat(daemon_name, fqdn_of_host);
 
     /* we need only SMTP for this purpose */
     if ((sock = SockOpen("localhost", SMTP_PORT, NULL, NULL)) == -1)
@@ -379,8 +422,11 @@ static int send_bouncemail(struct query *ctl, struct msgblk *msg,
     SockPrintf(sock, "--%s\r\n", boundary); 
     SockPrintf(sock, "Content-Type: text/rfc822-headers\r\n");
     SockPrintf(sock, "\r\n");
-    SockWrite(sock, msg->headers, strlen(msg->headers));
-    SockPrintf(sock, "\r\n");
+    if (msg->headers)
+    {
+	SockWrite(sock, msg->headers, strlen(msg->headers));
+	SockPrintf(sock, "\r\n");
+    }
     SockPrintf(sock, "--%s--\r\n", boundary); 
 
     if (SMTP_eom(sock) != SM_OK || SMTP_quit(sock))
@@ -400,6 +446,8 @@ static int handle_smtp_report(struct query *ctl, struct msgblk *msg)
 {
     int smtperr = atoi(smtp_response);
     char *responses[1];
+    struct idlist *walk;
+    int found = 0;
 
     xalloca(responses[0], char *, strlen(smtp_response)+1);
     strcpy(responses[0], smtp_response);
@@ -424,7 +472,15 @@ static int handle_smtp_report(struct query *ctl, struct msgblk *msg)
      * messages, which are probably in English (none of the
      * MTAs I know about are internationalized).
      */
-    if (str_find(&ctl->antispam, smtperr))
+    for( walk = ctl->antispam; walk; walk = walk->next )
+        if ( walk->val.status.num == smtperr ) 
+	{ 
+		found=1;
+		break;
+	}
+
+    /* if (str_find(&ctl->antispam, smtperr)) */
+    if ( found )
     {
 	/*
 	 * SMTP listener explicitly refuses to deliver mail
@@ -505,10 +561,12 @@ static int handle_smtp_report(struct query *ctl, struct msgblk *msg)
     default:
 	/* bounce non-transient errors back to the sender */
 	if (smtperr >= 500 && smtperr <= 599)
-	    if (send_bouncemail(ctl, msg, XMIT_ACCEPT,
+	{
+	    send_bouncemail(ctl, msg, XMIT_ACCEPT,
 				"General SMTP/ESMTP error.\r\n", 
-				1, responses))
-		return(run.bouncemail ? PS_REFUSED : PS_TRANSIENT);
+				1, responses);
+	    return(PS_REFUSED);
+	}
 	/*
 	 * We're going to end up here on 4xx errors, like:
 	 *
@@ -584,7 +642,7 @@ int stuffline(struct query *ctl, char *buf)
     /* The line may contain NUL characters. Find the last char to use
      * -- the real line termination is the sequence "\n\0".
      */
-    last = buf;
+    last = buf + 1; /* last[-1] must be valid! */
     while ((last += strlen(last)) && (last[-1] != '\n'))
         last++;
 
@@ -618,10 +676,15 @@ int stuffline(struct query *ctl, char *buf)
 	}
         else /* if (!protocol->delimited)	-- not byte-stuffed already */
 	{
-	    if (!ctl->mda)
-		SockWrite(ctl->smtp_socket, buf, 1);	/* byte-stuff it */
-	    else
-		/* leave it alone */;
+	  if (!ctl->mda)      /* byte-stuff it */
+	    {
+	      if (!ctl->bsmtp)
+		SockWrite(ctl->smtp_socket, buf, 1);
+	      else
+		{
+		  fwrite(buf, 1, 1, sinkfp);
+		}
+	    }
 	}
     }
 
@@ -767,6 +830,7 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
 #endif /* EXPLICIT_BOUNCE_ON_BAD_ADDRESS */
     int		total_addresses;
     int		force_transient_error = 0;
+    int		smtp_err;
 
     /*
      * Compute ESMTP options.
@@ -811,7 +875,7 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
      * Also, if the hostname is a dotted quad, wrap it in square brackets.
      * Apparently this is required by RFC2821, section 4.1.3.
      */
-    if (!msg->return_path[0] || (0 == strcmp(msg->return_path, "@")))
+    if (!msg->return_path[0] || (msg->return_path[0] == '@'))
     {
       if (is_dottedquad(ctl->server.truename))
       {
@@ -835,6 +899,9 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
     }
     else if (strchr(msg->return_path,'@') || strchr(msg->return_path,'!'))
 	ap = msg->return_path;
+    /* in case Return-Path was "<>" we want to preserve that */
+    else if (strcmp(msg->return_path,"<>") == 0)
+	ap = msg->return_path;
     else		/* in case Return-Path existed but was local */
     {
       if (is_dottedquad(ctl->server.truename))
@@ -858,7 +925,12 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
 	ap = addr;
     }
 
-    if (SMTP_from(ctl->smtp_socket, ap, options) != SM_OK)
+    if ((smtp_err = SMTP_from(ctl->smtp_socket, ap, options)) == SM_UNRECOVERABLE)
+    {
+	smtp_close(ctl, 0);
+	return(PS_TRANSIENT);
+    }
+    if (smtp_err != SM_OK)
     {
 	int err = handle_smtp_report(ctl, msg);
 
@@ -880,7 +952,12 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
 	{
 	    const char *address;
 	    address = rcpt_address (ctl, idp->id, 1);
-	    if (SMTP_rcpt(ctl->smtp_socket, address) == SM_OK)
+	    if ((smtp_err = SMTP_rcpt(ctl->smtp_socket, address)) == SM_UNRECOVERABLE)
+	    {
+		smtp_close(ctl, 0);
+		return(PS_TRANSIENT);
+	    }
+	    if (smtp_err == SM_OK)
 		(*good_addresses)++;
 	    else
 	    {
@@ -951,8 +1028,13 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
 	    SMTP_rset(ctl->smtp_socket);	/* required by RFC1870 */
 	    return(PS_REFUSED);
 	}
-	if (SMTP_rcpt(ctl->smtp_socket,
-		rcpt_address (ctl, run.postmaster, 0)) != SM_OK)
+	if ((smtp_err = SMTP_rcpt(ctl->smtp_socket,
+		rcpt_address (ctl, run.postmaster, 0))) == SM_UNRECOVERABLE)
+	{
+	    smtp_close(ctl, 0);
+	    return(PS_TRANSIENT);
+	}
+	if (smtp_err != SM_OK)
 	{
 	    report(stderr, GT_("can't even send to %s!\n"), run.postmaster);
 	    SMTP_rset(ctl->smtp_socket);	/* required by RFC1870 */
@@ -967,7 +1049,12 @@ static int open_smtp_sink(struct query *ctl, struct msgblk *msg,
      * Tell the listener we're ready to send data.
      * Some listeners (like zmailer) may return antispam errors here.
      */
-    if (SMTP_data(ctl->smtp_socket) != SM_OK)
+    if ((smtp_err = SMTP_data(ctl->smtp_socket)) == SM_UNRECOVERABLE)
+    {
+	smtp_close(ctl, 0);
+	return(PS_TRANSIENT);
+    }
+    if (smtp_err != SM_OK)
     {
 	int err = handle_smtp_report(ctl, msg);
 	SMTP_rset(ctl->smtp_socket);    /* stay on the safe side */
@@ -987,9 +1074,6 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	      int *good_addresses, int *bad_addresses)
 /* open a stream to a local MDA */
 {
-#ifdef HAVE_SIGACTION
-    struct      sigaction sa_new;
-#endif /* HAVE_SIGACTION */
 #ifdef HAVE_SETEUID
     uid_t orig_uid;
 #endif /* HAVE_SETEUID */
@@ -1035,7 +1119,6 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	    names[--nameslen] = '\0';	/* chop trailing space */
 	}
 
-	/* sanitize names in order to contain only harmless shell chars */
 	sanitize(names);
     }
 
@@ -1044,7 +1127,6 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
     {
 	from = xstrdup(msg->return_path);
 
-	/* sanitize from in order to contain *only* harmless shell chars */
 	sanitize(from);
 
 	fromlen = strlen(from);
@@ -1058,17 +1140,17 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	/* find length of resulting mda string */
 	sp = before;
 	while ((sp = strstr(sp, "%s"))) {
-	    length += nameslen - 2;	/* subtract %s */
+	    length += nameslen;	/* subtract %s and add '' */
 	    sp += 2;
 	}
 	sp = before;
 	while ((sp = strstr(sp, "%T"))) {
-	    length += nameslen - 2;	/* subtract %T */
+	    length += nameslen;	/* subtract %T and add '' */
 	    sp += 2;
 	}
 	sp = before;
 	while ((sp = strstr(sp, "%F"))) {
-	    length += fromlen - 2;	/* subtract %F */
+	    length += fromlen;	/* subtract %F and add '' */
 	    sp += 2;
 	}
 
@@ -1081,13 +1163,17 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
 	    /* need to expand? BTW, no here overflow, because in
 	    ** the worst case (end of string) sp[1] == '\0' */
 	    if (sp[1] == 's' || sp[1] == 'T') {
+		*dp++ = '\'';
 		strcpy(dp, names);
 		dp += nameslen;
+		*dp++ = '\'';
 		sp++;	/* position sp over [sT] */
 		dp--;	/* adjust dp */
 	    } else if (sp[1] == 'F') {
+		*dp++ = '\'';
 		strcpy(dp, from);
 		dp += fromlen;
+		*dp++ = '\'';
 		sp++;	/* position sp over F */
 		dp--;	/* adjust dp */
 	    }
@@ -1142,14 +1228,7 @@ static int open_mda_sink(struct query *ctl, struct msgblk *msg,
      * sigchld_handler() would reap away the error status, returning
      * error status instead of 0 for successful completion.
      */
-#ifndef HAVE_SIGACTION
-    signal(SIGCHLD, SIG_DFL);
-#else
-    memset (&sa_new, 0, sizeof sa_new);
-    sigemptyset (&sa_new.sa_mask);
-    sa_new.sa_handler = SIG_DFL;
-    sigaction (SIGCHLD, &sa_new, NULL);
-#endif /* HAVE_SIGACTION */
+    set_signal_handler(SIGCHLD, SIG_DFL);
 
     return(PS_SUCCESS);
 }
@@ -1237,6 +1316,7 @@ void release_sink(struct query *ctl)
 int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 /* perform end-of-message actions on the current output sink */
 {
+    int smtp_err;
     if (ctl->mda)
     {
 	int rc;
@@ -1290,7 +1370,12 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
     else if (forward)
     {
 	/* write message terminator */
-	if (SMTP_eom(ctl->smtp_socket) != SM_OK)
+	if ((smtp_err = SMTP_eom(ctl->smtp_socket)) == SM_UNRECOVERABLE)
+	{
+	    smtp_close(ctl, 0);
+	    return(FALSE);
+	}
+	if (smtp_err != SM_OK)
 	{
 	    if (handle_smtp_report(ctl, msg) != PS_REFUSED)
 	    {
@@ -1352,7 +1437,12 @@ int close_sink(struct query *ctl, struct msgblk *msg, flag forward)
 		xalloca(responses, char **, sizeof(char *) * lmtp_responses);
 		for (errors = i = 0; i < lmtp_responses; i++)
 		{
-		    if (SMTP_ok(ctl->smtp_socket) == SM_OK)
+		    if ((smtp_err = SMTP_ok(ctl->smtp_socket)) == SM_UNRECOVERABLE)
+		    {
+			smtp_close(ctl, 0);
+			return(FALSE);
+		    }
+		    if (smtp_err == SM_OK)
 			responses[i] = (char *)NULL;
 		    else
 		    {
@@ -1424,7 +1514,7 @@ int open_warning_by_mail(struct query *ctl, struct msgblk *msg)
     }
     else				/* send to postmaster  */
 	status = open_sink(ctl, &reply, &good, &bad);
-    stuff_warning(ctl, "Date: %s", rfc822timestamp());
+    if (status == 0) stuff_warning(ctl, "Date: %s", rfc822timestamp());
     return(status);
 }
 
@@ -1438,7 +1528,8 @@ va_dcl
 #endif
 /* format and ship a warning message line by mail */
 {
-    char	buf[POPBUFSIZE];
+    /* make huge -- i18n can bulk up error messages a lot */
+    char	buf[2*MSGBUFSIZE+4];
     va_list ap;
 
     /*
@@ -1465,13 +1556,18 @@ va_dcl
     strcat(buf, "\r\n");
 #endif /* HAVE_SNPRINTF */
 
+    /* guard against very long lines */
+    buf[MSGBUFSIZE+1] = '\r';
+    buf[MSGBUFSIZE+2] = '\n';
+    buf[MSGBUFSIZE+3] = '\0';
+
     stuffline(ctl, buf);
 }
 
 void close_warning_by_mail(struct query *ctl, struct msgblk *msg)
 /* sign and send mailed warnings */
 {
-    stuff_warning(ctl, GT_("--\r\n\t\t\t\tThe Fetchmail Daemon\r\n"));
+    stuff_warning(ctl, GT_("--\n\t\t\t\tThe Fetchmail Daemon\n"));
     close_sink(ctl, msg, TRUE);
 }
 
