@@ -726,7 +726,7 @@ int main(int argc, char *argv[])
      if (buf_p > buffer) {
         if (bodytype & MSG_NEEDS_DECODE) {
            buf_p = buffer;
-           UnMimeBodyline(&buf_p, 0);
+           UnMimeBodyline(&buf_p, 0, 0);
         }
         fwrite(buffer, (buf_p - buffer), 1, stdout);
         DBG_FWRITE(buffer, (buf_p - buffer), 1, fd_conv);
